#!/usr/bin/python
# -*- coding: utf-8 -*-

# attention: pour les accents, ce fichier doit être impérativement
#            lu et modifié sous éditeur avec encodage UTF-8

################################################################
######                                                    ######
######       Module additionnel: combinatoire.py          ######
######                                                    ######
######            accompagnant la CALCULEXT               ######
######      [Calculatrice Extensible Version 1.40]        ######
######                                                    ######
################################################################

# Copyright (C) 2008 Jean-Paul Vidal dit "Tyrtamos" — Tous droits réservés.
#   Logiciel distribué sur le site <http://linux.jpvweb.com>
#
#   Ce programme est un logiciel libre. Vous pouvez le redistribuer ou le
#   modifier suivant les termes de la "GNU General Public License" version 3,
#   telle que publiée par la Free Software Foundation.
#
#   Ce programme est distribué dans l’espoir qu’il vous sera utile, mais SANS
#   AUCUNE GARANTIE : sans même la garantie implicite de COMMERCIALISABILITÉ
#   ni d’ADÉQUATION À UN OBJECTIF PARTICULIER. Consultez la Licence Générale
#   Publique GNU pour plus de détails.
#
#   Ce programme doit être distribué avec une copie de la
#   License Générale Publique GNU (fichier license.html).
#   Si ce n’est pas le cas, consultez: <http://www.gnu.org/licenses/gpl.html>.

##################################################################
# Attention: configuration volontairement inhabituelle pour python
# les divisions entre 2 entiers seront décimales et pas entières!
# par exemple: 10/3 donnera "3.33333333333333" et pas "3"
# (les divisions entières seront désormais faites avec "//")
#
from __future__ import division

##################################################################
# Importation des modules nécessaires aux fonctions du présent module
#
from sys import modules
pmain = modules['__main__']

from math import *

###############################################################################
# Nombre de permutations de n objets en tenant compte de l'ordre
# NB: par convention: permut(0)=1 (mais pas de sens physique)
# - version volontairement non récursive (la pile de python est limitée à environ 1000)
#
def permut(n):
    """permut(n): nombre de permutations de n objets en tenant compte de l'ordre"""
    # Vérification de la donnée
    if (not pmain.estentier(n)) or (n<0):
        raise ValueError ("erreur permut(n): n doit être un nombre entier positif ou nul")
    # Calcul
    x=1
    for i in xrange(2,n+1):
        x*=i
    return x

###############################################################################
# Donne la liste de toutes les permutations d'une liste de chaîne
#   si r=1 => élimine les redondances dans le résultat
#
def permutliste(L,r=0):
    """permutliste(L,r=0): donne la liste de toutes les permutations d'une liste de chaîne (si r=1 => pas de redondance)"""
    def _permutliste(L,k=0):
        lg=len(L)
        if k==(lg-1):
            if (r==0) or ((r!=0) and (not(L in R))):
                R.append(L)
        else:
            for i in range(k,lg):
                Lt = L[0:k] + L[k+1:]
                Lt.append(L[k])
                L=Lt
                _permutliste(L,k+1)
    # Vérification de la donnée
    if (not pmain.estliste(L)) or (L==[]):
        raise ValueError ("erreur permutliste(L,r=0): L doit être une liste non vide")
    if (not pmain.estnombre(r)) or (not pmain.estentier(r)) or (r<0):
        raise ValueError ("erreur permutliste(L,r=0): r doit être un nombre entier positif ou nul")
    # Calcul
    R=[]
    _permutliste(L)
    return R

###############################################################################
# Donne la liste de toutes les permutations des caractères d'une chaîne de caractères
#   si r=1 => élimine les redondances dans le résultat
#
def permutchaine(ch,r=0):
    """permutchaine(ch,r=0): donne la liste de toutes les permutations des caractères d'une chaîne de caractères (si r=1 => pas de redondance)"""
    # Vérification de la donnée
    if (not pmain.estchaine(ch)) or (ch==""):
        raise ValueError ("erreur permutchaine(ch,r=0): ch doit être une chaîne de caractère non vide")
    if (not pmain.estnombre(r)) or (not pmain.estentier(r)) or (r<0):
        raise ValueError ("erreur permutchaine(ch,r=0): r doit être un nombre entier positif ou nul")
    # Calcul
    L=list(ch)
    R=permutliste(L,r)
    return [''.join(X) for X in R]

###############################################################################
# Calcul du nombre d'arrangements de n objets pris k a k
#
def arrang(n,k):
    """arrang(n,k): Calcul du nombre d'arrangements de n objets pris k a k"""
    # Vérification des données
    if (not pmain.estentier(n)) or (n<=0):
        raise ValueError ("erreur arrang(n,k): n doit être un nombre entier positif")
    if (not pmain.estentier(k)) or (k<0):
        raise ValueError ("erreur arrang(n,k): k doit être un nombre entier positif ou nul")
    if (k>n):
        raise ValueError ("erreur arrang(n,k): k doit être inférieur ou égal à n")
    # Calcul
    x=1
    i=n-k+1
    while i<=n:
        x*=i
        i+=1
    return x

#################################################################################
# Donne la liste de tous les arrangements de n objets pris k à k
#  les objets étant donnés sous forme d'une liste de chaîne
#
def arrangliste(L,k):
    """arrangliste(L,k): donne la liste de tous les arrangements de n objets (=liste L) pris k à k"""
    # Vérification de la donnée
    if (not pmain.estliste(L)) or (L==[]):
        raise ValueError ("erreur arrangliste(L,k): L doit être une liste non vide")
    if (not pmain.estnombre(k)) or (not pmain.estentier(k)) or (k<0):
        raise ValueError ("erreur arrangliste(L,k): k doit être un nombre entier positif ou nul")
    # Calcul
    T=combinliste(L,k)
    R=[]
    for X in T:
        R+=permutliste(X)
    return R

#################################################################################
# donne la liste de tous les arrangements de n objets (=caractères de la chaîne ch) pris k à k
#
def arrangchaine(ch,k):
    """arrangchaine(ch,k): donne la liste de tous les arrangements de n objets (=caractères de la chaîne ch) pris k à k"""
    # Vérification des données
    if (not pmain.estchaine(ch)) or (ch==""):
        raise ValueError ("erreur arrangchaine(ch,k): ch doit être une chaîne de caractère non vide")
    if (not pmain.estnombre(k)) or (not pmain.estentier(k)) or (k<0):
        raise ValueError ("erreur arrangchaine(ch,k): k doit être un nombre entier positif ou nul")
    # Calcul
    L=list(ch)
    R=arrangliste(L,k)
    return [''.join(X) for X in R]

###############################################################################
# Combinaison de n objets pris k a k
# NB: par convention, comb(n,0)=1 (mais pas de sens physique)
#
def combin(n,k):
    """combin(n,k): combinaison de n objets pris k a k"""
    # Vérification des données
    if (not pmain.estentier(n)) or (n<=0):
        raise ValueError ("erreur combin(n,k): n doit être un nombre entier positif")
    if (not pmain.estentier(k)) or (k<0):
        raise ValueError ("erreur combin(n,k): n doit être un nombre entier positif ou nul")
    if (k>n):
        raise ValueError ("erreur combin(n,k): k doit être inférieur ou égal à n")
    # Calcul
    x=1
    for i in xrange(n-k+1,n+1):
        x*=i
    y=1
    for i in xrange(2,k+1):
        y*=i
    return x//y

#################################################################################
# Donne la liste de toutes les combinaisons de n objets pris k à k
#  les objets étant donnés sous forme d'une liste de chaîne
#
def combinliste(L,k):
    """combinliste(L,k): donne la liste de toutes les combinaisons de n objets (=liste L) pris k à k"""
    # Vérification de la donnée
    if (not pmain.estliste(L)) or (L==[]):
        raise ValueError ("erreur combinliste(L,k): L doit être une liste non vide")
    if (not pmain.estnombre(k)) or (not pmain.estentier(k)) or (k<0):
        raise ValueError ("erreur combinliste(L,k): k doit être un nombre entier positif ou nul")
    # Calcul
    n=len(L)
    ch='1'*(k) + '0'*(n-k)
    T=permutchaine(ch,1)
    R=[]
    for car in T:
        X=[]
        for j in range(0,len(car)):
            if car[j]=='1':
                X.append(L[j])
        R.append(X)
    return R

#################################################################################
def combinchaine(ch,k):
    """combinchaine(ch,k): donne la liste de toutes les combinaisons de n objets (=caractères de la chaîne ch) pris k à k"""
    # Vérification des données
    if (not pmain.estchaine(ch)) or (ch==""):
        raise ValueError ("erreur combinchaine(ch,k): ch doit être une chaîne de caractère non vide")
    if (not pmain.estnombre(k)) or (not pmain.estentier(k)) or (k<0):
        raise ValueError ("erreur combinchaine(ch,k): k doit être un nombre entier positif ou nul")
    # Calcul
    L=list(ch)
    R=combinliste(L,k)
    return [''.join(X) for X in R]

